(** Defines classes for the IQ, Presence, and Message Jabber packets. *)

open Yaxpodom

exception Bad_packet

(** The base type of packet objects. *)
class type packet_base =
object
  method get_from : (unit -> string option)
  method set_from : (string option -> unit)

  method get_to : (unit -> string option)
  method set_to : (string option -> unit)

  method get_id : (unit -> string option)
  method set_id : (string option -> unit)

  method get_children : (unit -> content list)
  method set_children : (content list -> unit)

  (** Returns the (code,description) from the [error] child of the element.
      
    @raise Not_found if there is no [error] child. *)
  method get_error : (unit -> txt*txt)

  (** Emits a an element from this packet. *)
  method to_ele : (unit -> ele)
end

(** iq "type" attribute values *)
type iq_type =
    Get
  | Set
  | Result
  | IQError

(** The type of IQ packet objects *)
class type iq =
object
  inherit packet_base

  (** @raise Bad_packet if the [type] specified in the XML is not recognized *)
  method get_type : (unit -> iq_type option)
  method set_type : (iq_type option -> unit)
    
  (** Returns the first [query] child element of the IQ.
    
    @raise Not_found if there is no query. *)
  method get_first_query : (unit -> ele)

  (** Returns the default XML namespace URI of the first [query] child element of the IQ.

    @raise Not_found if there is no such element. *)
  method get_first_query_xmlns : (unit -> string)
end

(** Create a blank IQ object given the prefix to use for jabber:client *)
val mk_iq : (string -> iq)

(** presence "type" attribute values *)
type presence_type =
    Unavailable
  | Subscribe
  | Subscribed
  | Unsubscribe
  | Unsubscribed
  | Probe
  | PresError

(** the type of presence packet objects *)
class type presence =
object
  inherit packet_base

  (** @raise Bad_packet if the [type] specified in the XML is not recognized *)
  method get_type : (unit -> presence_type option)
  method set_type : (presence_type option -> unit)

  method get_show : (unit -> string option)
  method get_status : (unit -> string option)
  method get_priority : (unit -> string option)
end

(** Create a blank presence object given the prefix to use for jabber:client *)
val mk_presence : (string -> presence)

(** message "type" attribute values *)
type message_type =
    Normal
  | Chat
  | Groupchat
  | Headline
  | Error

(** the type of message packet objects *)
class type message =
object
  inherit packet_base

  (** @raise Bad_packet if the [type] specified in the XML is not recognized *)
  method get_type : (unit -> message_type option)
  method set_type : (message_type option -> unit)

  method get_body : (unit -> string option)
  method get_subject : (unit -> string option)
  method get_thread : (unit -> string option)
end

(** create a blank message object given the prefix to use for jabber:client *)
val mk_message : (string -> message)

(** The union type of packets. *)
type packet =
    IQ of iq
  | Presence of presence
  | Message of message

(** Construct a packet from an element

@raise Bad_packet if ele is not of jabber:client *)
val packet_of_ele : (ele -> packet)

val packet_base_of_packet : (packet -> packet_base)
val packet_from : (packet -> string option)
val packet_to : (packet -> string option)
val packet_id : (packet -> string option)
(** Generic accessors. *)
