open Xmlstream
open Yaxpodom
open Jabber
open Jabbersm
open Unix

let out = function
    Some(x) -> x
  | None -> assert false

type state =
    New
  | Connected
  | Authenticated
  | Ended
  | Failed

exception Protocol_error of string
exception Auth_error of string

class type session =
object
  method connect : (string -> int -> unit)
  method get_stream_id : (unit -> string)
  method authenticate : (Auth.credentials -> string -> unit)
  method pull_packet : (unit -> Jabber.packet)
  method write_packet : (Jabber.packet -> unit)
  method close : (unit -> unit)
end

class session_impl =
object (self)
  val mutable my_state = New
  val mutable s_in = None
  val mutable s_out = None
  val reader = mk_xml_stream_reader ()
  val writer = mk_xml_stream_writer ()
  val inbuf = String.create 256
  val stream_id = ref None

  method read_some () =
    let len = input (out s_in) inbuf 0 256 in
      reader#add_data inbuf 0 len

  method do_write () =
    writer#write_fully
      (fun s ofs len ->
	 output (out s_out) s ofs len);
    flush (out s_out)

  method pull_handshake () =
    (try
       reader#pull_handshake ()
     with Not_enough -> self#read_some(); self#pull_handshake ())

  method run_machine m =
    m#add_packet (self#pull_packet ());
    (match m#get_state () with
	 Accept -> Accept
       | Reject -> Reject
       | _ -> self#run_machine m)

  method connect server port =
    if not (my_state = New) then
      failwith "unexpected connect";
    let addrs = (gethostbyname server).h_addr_list in
    let (s_in',s_out') = open_connection (ADDR_INET (addrs.(0),port)) in
      s_in <- Some s_in';
      s_out <- Some s_out';
      writer#handshake
	"http://etherx.jabber.org/streams" "stream" "jabber:client"
	[ Att(QName("","to",""), ref server) ];
      self#do_write ();
      let (name,atts) = self#pull_handshake () in
	(match name with
	     QName(_,"stream","http://etherx.jabber.org/streams") -> ()
	   | _ -> raise (Protocol_error "Unrecognized stream"));
	List.iter
	  (function
	       Att(QName("","id",""),the_id) -> stream_id := Some(!the_id)
	     | _ -> ()) atts;
	if !stream_id = None then
	  raise (Protocol_error "Server did not specify stream ID");
	my_state <- Connected

  method get_stream_id () =
    (match my_state with
	 New -> failwith "unexpected get_stream_id"
       | _ -> out !stream_id)

  method authenticate credentials resource =
    if not (my_state = Connected) then
      failwith "Unexpected authenticate";
    let auth_machine = Auth.mk_auth_machine credentials resource
			 (fun p -> self#write_packet p) in
      auth_machine#start ();
      (match self#run_machine auth_machine with
	   Accept -> my_state <- Authenticated
	 | Reject -> raise (Auth_error (auth_machine#reject_reason ()))
	 | _ -> assert false)

  method pull_packet () =
    if (my_state = New) || (my_state = Ended) || (my_state = Failed) then
      failwith "unexpected pull_packet";
    (try
       packet_of_ele (reader#pull_ele ())
     with
	 Not_enough -> self#read_some (); self#pull_packet ())

  method write_packet p =
    if (my_state = New) || (my_state = Ended) || (my_state = Failed) then
      failwith "unexpected write_packet";
    let p' = packet_base_of_packet p in
      writer#add_ele (p'#to_ele ()) true;
      self#do_write ()

  method close () =
    if not ((my_state = Connected) || (my_state = Authenticated)) then
      failwith "unexpected close";
    writer#end_stream ();
    self#do_write ();
    close_out (out s_out);
    close_in (out s_in);
    my_state <- Ended
end

let mk_session () =
  ((new session_impl) :> session)
