(** Synchronous (blocking) Jabber client sessions, suitable for creating components and bots. You must link with the [nums], [unix], and [cryptokit] in addition to the [jabber] and [jabber_sync] modules to use this. *)

(** Raised if a protocol error occurs, e.g., unrecognized stream namespace from server *)
exception Protocol_error of string

(** Raised if authentication fails. *)
exception Auth_error of string

(** The type of a Jabber session object. *)
class type session =
object

  (** [s#connect server port] establishes an XML stream with the specified server on the specified port. 
    
    @raise Unix.Unix_error if a connection error occurs
    @raise Protocol_error if there was some problem with the server's XML stream. *)
  method connect : (string -> int -> unit)

  (** Retrieves the stream ID set by the server after a successful [connect], which may be needed to build credentials for [authenticate]. *)
  method get_stream_id : (unit -> string)

  (** Authenticates with the server using the specified credentials and resource.
    
    @raise Auth_error if the server rejects the credentials or resource. *)
  method authenticate : (Auth.credentials -> string -> unit)

  (** Waits for the next packet from the server.
    
    @raise Xmlstream.Stream_end if the server ends the stream.
    @raise Xmlstream.Stream_error if the server sends a stream error. *)
  method pull_packet : (unit -> Jabber.packet)

  (** Writes a packet to the server. *)
  method write_packet : (Jabber.packet -> unit)

  (** Closes the session. *)
  method close : (unit -> unit)
end

(** Creates a new session. *)
val mk_session : (unit -> session)
