type t = string option ref * string ref * string option ref

let out = function
    Some(x) -> x
  | None -> assert false

let mk_jid node domain resource = ((ref node,ref domain,ref resource):t)

exception Bad_jid

let is_node_char c' =
  let c = U_char.to_int c' in
    c >= 0x20 &&
    not (c = 0x22) &&
    not (c = 0x26) &&
    not (c = 0x27) &&
    not (c = 0x2F) &&
    not (c = 0x3A) &&
    not (c = 0x3C) &&
    not (c = 0x3E) &&
    not (c = 0x40) &&
    not (c = 0x7F) &&
    not (c = 0xFFFE) &&
    not (c = 0xFFFF)

let check_node str ofs len =
  let rec helper len' =
    if len' < len then
      let firstb = Char.code(String.get str (ofs + len')) in
      let csize = Utf8_char.utf8_character_size firstb in
      let echr = Utf8_char.of_bytes (String.sub str (ofs + len') csize) in
      let chr = Utf8_char.to_U_char echr in
	(if is_node_char chr then
	   helper (len' + csize)
	 else
	   raise Bad_jid) in
    helper 0

let is_domain_letter c =
  (c >= (Char.code 'a') &&
   c <= (Char.code 'z')) ||
  (c >= (Char.code 'A') &&
   c <= (Char.code 'Z'))

let is_domain_digit c =
  (c >= (Char.code '0') &&
   c <= (Char.code '9'))

let is_domain_other c =
  (c = (Char.code '.')) ||
  (c = (Char.code '-'))

let check_domain str ofs len =
  let rec helper len' =
    if len' < len then
      let nextc = (Char.code (String.get str (ofs + len'))) in
	if not ((is_domain_letter nextc) || (is_domain_digit nextc) || (is_domain_other nextc)) then
	  raise Bad_jid;
	helper (len' + 1 ) in
    helper 0

let check_resource str ofs len =
  let rec helper len' =
    if len' < len then
      let firstb = Char.code(String.get str (ofs + len')) in
      let csize = Utf8_char.utf8_character_size firstb in
      let echr = Utf8_char.of_bytes (String.sub str (ofs + len') csize) in
      let chr = Utf8_char.to_U_char echr in
      let code = U_char.to_int chr in
	(if not (code = 0xFFFE || code = 0xFFFF)  then
	   helper (len' + csize)
	 else
	   raise Bad_jid) in
    helper 0

let of_string str =
  if (String.length str) = 0 then
    raise Bad_jid;
  try
    (try
       let atpos = String.index str '@' in
	 if not (atpos > 0 && (atpos + 1) < (String.length str)) then
	   raise Bad_jid;
	 (try
	    let slashpos = String.index str '/' in
	      if slashpos < (atpos + 1) then
		raise Bad_jid;
	      check_node str 0 atpos;
	      check_domain str (atpos + 1) (slashpos - atpos - 1);
	      check_resource str (slashpos + 1) ((String.length str) - slashpos - 1);
	      (ref (Some (String.sub str 0 atpos)),
	       ref (String.sub str (atpos + 1) (slashpos - atpos - 1)),
	       ref (Some (String.sub str (slashpos + 1) ((String.length str - slashpos - 1)))))
	  with
	      Not_found ->
		check_node str 0 atpos;
		check_domain str (atpos + 1) ((String.length str) - atpos - 1);
		(ref (Some (String.sub str 0 atpos)),
		 ref (String.sub str (atpos + 1) ((String.length str) - atpos - 1)),
		 ref None))
     with
	 Not_found ->
	   (try
	      let slashpos = String.index str '/' in
		if slashpos < 1 then
		  raise Bad_jid;
		check_domain str 0 slashpos;
		check_resource str (slashpos + 1) ((String.length str) - slashpos - 1);
		(ref None,
		 ref (String.sub str 0 slashpos),
		 ref (Some (String.sub str (slashpos + 1) ((String.length str - slashpos - 1)))))
	    with
		Not_found ->
		  check_domain str 0 (String.length str);
		  ((ref None,ref str,ref None):t)))
  with
      U_char.Bad_char | Utf8_char.Bad_encoding | Invalid_argument(_) -> raise Bad_jid

let to_string ((node,domain,resource):t) =
  (match (!node,!resource) with
       (Some n,Some r) -> n ^ "@" ^ !domain ^ "/" ^ r
     | (Some n,None) -> n ^ "@" ^ !domain
     | (None,Some r) -> !domain ^ "/" ^ r
     | (None,None) -> !domain)

let node ((n,d,r):t) = !n
let domain ((n,d,r):t) = !d
let resource ((n,d,r):t) = !r

let set_node ((n,d,r):t) n' = n := n'
let set_domain ((n,d,r):t) d' = d := d'
let set_resource ((n,d,r):t) r' = r := r'

let strip_resource ((n,d,r):t) = ((n,d,ref None):t)

