open Yaxpodom

type reader_state =
    Going
  | Pausing
  | Paused of (unit->unit)

(* TODO: redo buffering so it's not stupid *)
class push_utf8_reader =
object (self)
  val mutable data = ""
  val mutable len = 0
  val mutable pos = 0
  val mutable state = Going

  method add_data s ofs len' =
    len <- len + len';
    let b = Buffer.create len in
      Buffer.add_string b data;
      Buffer.add_substring b s ofs len';
      data <- Buffer.contents b

  method pause () =
    (match state with
	 Going -> state <- Pausing
       | _ -> ())

  method is_paused () =
    (match state with
	 Going -> false
       | _ -> true)

  method resume () =
    (match state with
	 Paused(k) -> state <- Going; k ()
       | _ -> state <- Going)

  method peek_char k =
    (match state with
	 Going ->
	   if pos < len then
	     let firstb = String.get data pos in
	     let size = Utf8_char.utf8_character_size (Char.code firstb) in
	       if pos + size <= len then
		 let chr = String.sub data pos size in
		   k (Utf8_char.of_bytes chr)
	       else
		  state <- (Paused (fun () -> self#peek_char k))
	   else
	     state <- (Paused (fun () -> self#peek_char k))
       | Pausing -> state <- (Paused (fun () -> self#peek_char k))
       | _ -> assert false)

  method read_char k = 
    self#peek_char (fun chr ->
		      pos <- pos + (Utf8_char.size chr);
		      if pos = len then
			(data <- "";
			 pos <- 0;
			 len <- 0);
		      k chr)

  method read_while pred k =
    let rec helper curpos =
      if state = Pausing || curpos = len then
	state <- (Paused (fun () -> helper curpos))
      else
	(assert (curpos < len);
	 let firstb = String.get data curpos in
	 let size = Utf8_char.utf8_character_size (Char.code firstb) in
	   if curpos + size <= len then
	     let chr = String.sub data curpos size in
	       if pred (Utf8_char.of_bytes chr) then
		 helper (curpos + size)
	       else
		 let rslt = String.sub data pos (curpos - pos) in
		   pos <- curpos;
		   if pos = len then
		     (data <- "";
		      pos <- 0;
		      len <- 0);
		   k rslt
	   else
	     state <- (Paused (fun () -> helper curpos))) in
      helper pos
end

let is_xmlns_prefix_decl' yaxpoatt =
  yaxpoatt.Yaxpo.att_name.Yaxpo.pfx = "xmlns"

let is_xmlns_default_decl' yaxpoatt =
  yaxpoatt.Yaxpo.att_name.Yaxpo.pfx = "" && yaxpoatt.Yaxpo.att_name.Yaxpo.local = "xmlns"

let is_xmlns_decl' yaxpoatt =
  (is_xmlns_prefix_decl' yaxpoatt) || (is_xmlns_default_decl' yaxpoatt)

let qname_of_yaxpo name uri =
  QName(name.Yaxpo.pfx,name.Yaxpo.local,uri)

let att_of_yaxpo att =
  Att((qname_of_yaxpo att.Yaxpo.att_name ""), ref att.Yaxpo.att_value)

let lookup_prefix prefix decls =
  att_value
    (List.find (fun (Att(QName(pfx,local,nsURI),_)) ->
		  (match (pfx,local) with
		       ("","xmlns") when prefix = "" -> true
		     | ("xmlns",prefix') when (not (prefix = "")) && (prefix=prefix') -> true
		     | _ -> false))	       
       decls)

let att_of_yaxpo' att decls =
  match att.Yaxpo.att_name with
      { Yaxpo.pfx = ""; Yaxpo.local = "xmlns" }
    | { Yaxpo.pfx = "xmlns"; Yaxpo.local = _ }
    | { Yaxpo.pfx = ""; Yaxpo.local = _ } -> att_of_yaxpo att
    | _ ->
	Att((qname_of_yaxpo att.Yaxpo.att_name
	       (lookup_prefix att.Yaxpo.att_name.Yaxpo.pfx decls)), ref att.Yaxpo.att_value)

type xml_stream_state =
    New
  | Handshaking
  | Normal
  | Ended
  | Failed

exception Not_enough
exception Stream_end
exception Stream_error of string

class type xml_stream_reader =
object
  method add_data : (string->int->int->unit)

  method get_state : (unit->xml_stream_state)

  method pull_handshake : (unit->(qname*att list))
  method pull_ele : (unit->ele)

  method stream_ns_URI : (unit->string)
  method stream_ns_pfx : (unit->string) 
  method default_ns_URI : (unit->string)
  method stream_ns_decls : (unit->att list)
end

class xml_stream_reader_impl =
object (self)
  val reader = new push_utf8_reader
  val mutable state = New
  val mutable yaxpo_ns_decls = []
  val mutable ns_decls = []
  val mutable stream_namespace_pfx = ""
  val mutable callbacks = None
  val mutable next_ele = None
  val mutable docele_tag_name = None
  val mutable docele_atts = None

  initializer
    callbacks <-
    (Some
       {
	 Yaxposax.start_element =
	  (fun name yaxpoatts kont ->
	     assert (state = Handshaking);
	     yaxpo_ns_decls <- List.filter is_xmlns_decl' yaxpoatts;
	     ns_decls <- List.map att_of_yaxpo yaxpo_ns_decls;
	     stream_namespace_pfx <- name.Yaxpo.pfx;
	     try
	       docele_atts <- Some (List.map (fun att -> att_of_yaxpo' att ns_decls) yaxpoatts);
	       docele_tag_name <- Some (qname_of_yaxpo name
					  (lookup_prefix stream_namespace_pfx ns_decls));
	       state <- Normal
	     with
		 Not_found ->
		   state <- Failed;
		   raise Namespace_error);
	 Yaxposax.characters = (fun str kont -> assert false);
	 Yaxposax.cdata = (fun data kont -> assert false);
	 Yaxposax.comment = (fun str kont -> kont ());
	 Yaxposax.pi = (fun target value kont -> kont ());
	 Yaxposax.end_element = (fun name kont -> assert false);
       })

  method get_state () =
    state

  method add_data s ofs len =
    reader#add_data s ofs len

  method pull_handshake () =
    match state with
	New ->
	  state <- Handshaking;
	  (try
	     (Yaxposax.parse_document reader
		(match callbacks with Some(c) -> c | _ -> assert false)
		(fun () -> assert false);
	      if state = Normal then
		(match (docele_tag_name,docele_atts) with
		     (Some(name),Some(atts)) -> (name,atts)
		   | _ -> assert false)
	      else
		raise Not_enough)
	   with
	       Not_enough -> raise Not_enough
	     | _ as exn -> state <- Failed; raise exn)
      | Handshaking ->
	  assert (reader#is_paused ());
	  (try
	     (reader#resume ();
	      if state = Normal then
		(match (docele_tag_name,docele_atts) with
		     (Some(name),Some(atts)) -> (name,atts)
		   | _ -> assert false)
	      else
		raise Not_enough)
	   with
	       Not_enough -> raise Not_enough
	     | _ as exn -> state <- Failed; raise exn)
      | _ -> failwith "unexpected pull_handshake"

  method pull_ele () =
    match state with
	Normal ->
	  (try
	     next_ele <- None;
	     if reader#is_paused () then
	       reader#resume ()
	     else
	       (let s = Stack.create () in
		  Stack.push yaxpo_ns_decls s;
		  build_element_tree_cps reader s
		    (fun the_ele ->
		       next_ele <- Some(the_ele)));
	     (match next_ele with
		  Some(
		    Element(
		      QName(_,"error",aURI),_,_) as the_ele)
		    when aURI = self#stream_ns_URI () ->
		      raise (Stream_error (ele_inner_text the_ele))
		| Some(e) -> e
		| None -> raise Not_enough)
	   with
	       Not_enough -> raise Not_enough
	     | Yaxpo.Parse_error
		 (_,Some(
		    Yaxpo.Tok_end_tag(
		      { Yaxpo.pfx = pfx; Yaxpo.local = local})))
		 when pfx = stream_namespace_pfx && local = "stream" ->
		 raise Stream_end
	     | _ as exn -> state <- Failed; raise exn)
      | _ -> failwith "unexpected pull_ele"

  method stream_ns_URI () =
    (match state with
	 New | Handshaking -> failwith "unexpected call to stream_ns_URI"
       | _ -> lookup_prefix stream_namespace_pfx ns_decls)

  method stream_ns_pfx () =
    (match state with
	 New | Handshaking -> failwith "unexpected call to stream_ns_pfx"
       | _ -> stream_namespace_pfx)

  method default_ns_URI () =
    (match state with
	 New | Handshaking -> failwith "unexpected call to default_ns_URI"
       | _ -> (lookup_prefix "" ns_decls))

  method stream_ns_decls () =
    (match state with
	 New | Handshaking -> failwith "unexpected call to stream_ns_decls"
       | _ -> ns_decls)
end

let mk_xml_stream_reader () =
  let r = new xml_stream_reader_impl in
    (r :> xml_stream_reader)

class type xml_stream_writer =
object
  method data_bytes_avail : (unit -> int)
  method write_some : (int->(string->int->int->int)->int)
  method write_fully : ((string->int->int->unit)->unit)
  method handshake : (txt->txt->txt->att list->unit)
  method add_ele : (ele -> bool -> unit)
  method end_stream : (unit->unit)
end

let is_xmlns_prefix_decl = function
    Att(QName("xmlns",_,""),_) -> true
  | _ -> false

let is_xmlns_default_decl = function
    Att(QName("","xmlns",""),_) -> true
  | _ -> false

let is_xmlns_decl att =
  (is_xmlns_prefix_decl att) || (is_xmlns_default_decl att)

let rec lookup prefix decls =
  let rec lookup_in decls =
    (att_value
       (List.find
	  (function
	       (Att(QName("","xmlns",_),value)) when prefix = "" -> true
	     | (Att(QName("xmlns",prefix',""),_)) when prefix = prefix' -> true
	     | _ -> false)
	  decls)) in
  (match decls with
       [] -> raise Not_found
     | these :: rest -> (try lookup_in these with Not_found -> lookup prefix rest))

let rec collect_undeclared_namespaces ele decls =
  let from_name name decls allow_none =
    try
      (match name with
	   QName("",_,"") when allow_none -> []
	 | QName("",_,"") -> failwith "qname.nsURI must be specified"
	 | QName(pfx,_,nsURI) ->
	     let nsURI' = lookup pfx decls in
	       if not (nsURI = nsURI')  then
		   raise Not_found
	       else
		 [])
    with
	Not_found ->
	  if (qname_nsURI name) = "" then
	    failwith "qname.nsURI must be specified";
	  [(qname_pfx name,qname_nsURI name)] in
  let from_atts supplemented_decls =
    List.flatten
      (List.map (fun an_att ->
		   if not (is_xmlns_decl an_att) then
		     from_name (att_name an_att) supplemented_decls true
		   else
		     []) (ele_atts ele)) in
  let from_children supplemented_decls =
    List.flatten
      (List.map (function
		     Subele(e) -> collect_undeclared_namespaces e supplemented_decls
		   | _ -> []) (ele_children ele)) in
  let supplemented_decls = (List.filter is_xmlns_decl (ele_atts ele)) :: decls in
    List.flatten [(from_name (ele_name ele) supplemented_decls false); (from_atts supplemented_decls); (from_children supplemented_decls)]
	 

(* TODO: redo buffering so it's not stupid *)
class xml_stream_writer_impl =
object (self)
  val mutable data = ""
  val mutable len = 0
  val mutable pos = 0
  val mutable state = New
  val mutable ns_pfx = "";
  val mutable ns_URI = "";
  val mutable def_ns_URI = "";
  val mutable ns_decls = [];

  method data_bytes_avail () = 
    len - pos

  method write_some len' out =
    let real_len = min (self#data_bytes_avail ()) len' in
    let written = out data pos real_len in
      pos <- pos + written;
      if pos = len then
	(data <- "";
	 len <- 0;
	 pos <- 0);
      written

  method write_fully (out:(string->int->int->unit)) =
    ignore (self#write_some (self#data_bytes_avail ()) 
	      (fun s ofs len ->
		 out s ofs len;
		 len))

  method private add_data str ofs len' =
    let b = Buffer.create ((len - pos) + len') in
      Buffer.add_substring b data pos (len - pos);
      Buffer.add_substring b str ofs len';
      data <- Buffer.contents b;
      len <- (len - pos) + len';
      pos <- 0

  method handshake stream_ns_URI stream_ns_prefix default_ns_URI atts =
    if not (state = New) then
      failwith "unexpected handshake";
    if stream_ns_prefix = "" && (not (stream_ns_URI = default_ns_URI)) then
      failwith "Inconsistent namespaces specified";
    let stream_ns_decls =
      List.flatten
	[(if not (stream_ns_prefix = "") then
	    [Att(QName("xmlns",stream_ns_prefix,""),ref stream_ns_URI)]
	  else
	    []);
	 (if not (default_ns_URI = "") then
	    [Att(QName("","xmlns",""),ref default_ns_URI)]
	  else if (stream_ns_prefix = "") then
	    [Att(QName("","xmlns",""),ref stream_ns_URI)]
	  else
	    [])] in
    let b = Buffer.create 128 in
      Buffer.add_char b '<';
      Buffer.add_string b (string_of_qname (QName(stream_ns_prefix,"stream","")));
      Buffer.add_string b (string_of_atts (List.concat [stream_ns_decls; atts]));
      Buffer.add_char b '>';
      let s = Buffer.contents b in
	self#add_data s 0 (String.length s);
	ns_pfx <- stream_ns_prefix;
	ns_URI <- stream_ns_URI;
	def_ns_URI <- default_ns_URI;
	ns_decls <- (List.filter is_xmlns_decl (List.concat [stream_ns_decls; atts]));
	state <- Normal

  method add_ele the_ele (process_namespaces:bool) =
    (match state with
	 Normal ->
	   let ele_copy = make_ele (ele_name the_ele) in
	     set_ele_atts ele_copy (ele_atts the_ele);
	     set_ele_children ele_copy (ele_children the_ele);
	     if process_namespaces then
	       set_ele_atts ele_copy
		 (List.concat
		    [(List.map
			(fun (prefix,uri) ->
			   Att((if prefix = "" then
				  QName("","xmlns","")
				else
				  QName("xmlns",prefix,"")),
				ref uri))
			(collect_undeclared_namespaces the_ele [ns_decls]));
		     (ele_atts ele_copy)]);
	     let s = (string_of_ele ele_copy) in
	       self#add_data s 0 (String.length s)
       | _ -> failwith "unexpected add_ele")

  method end_stream () =
    let b = Buffer.create 16 in
      Buffer.add_string b "</";
      Buffer.add_string b (string_of_qname (QName(ns_pfx,"stream","")));
      Buffer.add_char b '>';
      let s = Buffer.contents b in
	self#add_data s 0 (String.length s);
	state <- Ended
end

let mk_xml_stream_writer () =
  let w = new xml_stream_writer_impl in
    (w :> xml_stream_writer )
