(** Provides XML streams, the basic Jabber wire protocol. *)

open Yaxpodom

(** State of an XML stream. *)
type xml_stream_state =
    New
  | Handshaking
  | Normal
  | Ended
  | Failed

(** Raised if not enough data has been received to complete a "pull" operation. *)
exception Not_enough

(** Raised if the server ends the stream. *)
exception Stream_end

(** Raised if the server sends a stream error. *)
exception Stream_error of string

(** The XML stream "reader". The reader accepts arbitrarily chunked data received from the server. It then allows you to "pull" elements from it; elements are parsed as you pull them. *)
class type xml_stream_reader =
object
  (** Add a chunk of data to the reader. [add_data s ofs len] adds [len] bytes from [s] starting at [ofs]. The data is not actually parsed until you pull something. *)
  method add_data : (string->int->int->unit)

  (** Returns the current state of the parser. *)
  method get_state : (unit->xml_stream_state)

  (** Pulls the XML Stream handshake, that is, the document start tag. Call this when data is first received from the server.

    @raise Not_enough if not enough data is available. In this case, receive more data from the server, and try again. *)
  method pull_handshake : (unit->(qname*att list))

  (** Pulls the next content element from the XML stream.
    
    @raise Not_enough if not enough data is available. In this case, receive more data from the server, and try again.
    @raise Stream_end if the server ends the stream.
    @raise Stream_error if the server sends a stream error.*)
  method pull_ele : (unit->ele)   

  (** {b The following methods are only meaningful after a successful [pull_handshake]:} *)

  (** Returns the toplevel namespace URI of the XML stream. *)
  method stream_ns_URI : (unit->string)

  (** Returns the prefix corresponding to the toplevel namespace URI. May be "" if the toplevel namespace URI is the default *)
  method stream_ns_pfx : (unit->string) 

  (** Returns the default namespace URI for the XML stream. *)
  method default_ns_URI : (unit->string)

  (** Returns the stream's toplevel namespace declarations. *)
  method stream_ns_decls : (unit->att list)
end

(** Creates an XML stream reader. *)
val mk_xml_stream_reader : (unit -> xml_stream_reader)

(** The XML stream writer. You give it elements, and then it lets you extract the serialized bytes to be written to the socket. *)
class type xml_stream_writer =
object
  (** Returns how many bytes are available to be transmitted *)
  method data_bytes_avail : (unit -> int)

  (** The procedure argument is expected to accept [str ofs len] and return how many bytes were successfully written. *)
  method write_some : (int->(string->int->int->int)->int)

  (** the procedure argument is expected to accept [str ofs len]. It should be prepared to be called multiple times. *)
  method write_fully : ((string->int->int->unit)->unit)

  (** [handshake stream_ns_uri stream_ns_pfx default_ns_URI atts] takes a stream (toplevel) namespace prefix and URI, default stream namespace URI, and list of attributes, and generates the stream start tag. The attributes MUST NOT contain the stream or default namespace declarations, but MAY contain additional declarations. *)
  method handshake : (txt->txt->txt->att list->unit)

  (** Adds an element to the stream. The boolean argument specfies whether the stream writer should perform namespace processing on the element. If it is [true], then the writer will walk the tree, determine which stream namespace prefixes are not declared inside the tree in the context of the rest of the stream, and add the appropriate declarations; this requires that all elements and attributes in the tree have the [nsURI] field of their [qname]s filled out. If the boolean argument is [false], then the element will just be put into the stream verbatim. This can result in an invalid XML stream. *)
  method add_ele : (ele -> bool -> unit)

  (** generates the stream end. *)
  method end_stream : (unit->unit)
end

(** Creates an XML stream writer. *)
val mk_xml_stream_writer : (unit->xml_stream_writer)
